/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.  
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are 
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1998-1999 Jeremie Miller.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 * 
 * --------------------------------------------------------------------------*/

#include "conference.h"
#include <sys/utsname.h>

void con_server_browsewalk(xht h, const char *key, void *data, void *arg)
{
    cnr room = (cnr)data;
    jpacket jp = (jpacket)arg;
    xmlnode x;
    char str[20];

    /* if we're a private server, we can only show the rooms the user already knows about */
    if(room->public == 0 && xhash_get(room->remote,jid_full(jp->to)) == NULL)
        return;

    x = xmlnode_insert_tag(jp->iq, "conference");

    if(room->public == 0)
        xmlnode_put_attrib(x, "type", "private");
    else
        xmlnode_put_attrib(x, "type", "public");

    xmlnode_put_attrib(x, "jid", jid_full(room->id));

    if(room->maxusers > 0)
	    sprintf(str, "%d/%d", room->count, room->maxusers);
    else
	    sprintf(str, "%d", room->count);

    xmlnode_put_attrib(x, "name", spools(jp->p, room->name, " (", str, ")", jp->p));
}

void con_server(cni master, jpacket jp)
{
    char *str;
    struct utsname un;
    xmlnode x;
    int start, status;
    time_t t;
    char *from;
    char nstr[10];

    log_debug(NAME, "[%s] server packet", FZONE);

    if(jp->type != JPACKET_IQ)
    {
        xmlnode_free(jp->x);
        return;
    }

    if(jpacket_subtype(jp) == JPACKET__SET)
    {
	if(NSCHECK(jp->iq,NS_REGISTER))
	{
            log_debug(NAME, "[%s] Server packet - Registration Submission", FZONE);
	    str = xmlnode_get_tag_data(jp->iq, "nickname");

	    from = xmlnode_get_attrib(jp->x, "from");
	    status = is_registered(master, from, str);

	    if(status == -1)
	    {
                log_debug(NAME, "[%s] Server packet - Registration Submission", FZONE);
		jutil_error(jp->x, TERROR_MUC_NICKREG);
	    }
	    else if(status == 0)
	    {
		set_data(master, str, from, NULL);
		jutil_iqresult(jp->x);
	    }
	    else
	    {
		jutil_iqresult(jp->x);
	    }

            deliver(dpacket_new(jp->x), NULL);
	    return;
	}
    }

    if(jpacket_subtype(jp) == JPACKET__GET)
    {
        /* now we're all set */

	if(NSCHECK(jp->iq,NS_REGISTER))
	{
            log_debug(NAME, "[%s] Server packet - Registration Request", FZONE);

	    x = get_data_byjid(master, xmlnode_get_attrib(jp->x, "from"));
	    str = xmlnode_get_attrib(x, "nick") ? xmlnode_get_attrib(x, "nick") : j_strdup("");
	    
	    jutil_iqresult(jp->x);
	    xmlnode_put_attrib(xmlnode_insert_tag(jp->x, "query"), "xmlns", NS_REGISTER);
	    jpacket_reset(jp);

            xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "instructions"), "Enter the nickname you wish to reserve for this conference service", -1);
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "nickname"), str, -1);
	    
            /* 
	     * xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "key"), jutil_regkey(NULL, jid_full(jp->to)), -1);
	     * if(x != NULL) xmlnode_insert_tag(jp->iq, "registered");
	     */
            deliver(dpacket_new(jp->x), NULL);
	    return;
													    
	}
	else if(NSCHECK(jp->iq,NS_TIME))
        {
            log_debug(NAME, "[%s] Server packet - Time Request", FZONE);

            jutil_iqresult(jp->x);
            xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"),"xmlns",NS_TIME);
            jpacket_reset(jp);
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"utc"),jutil_timestamp(),-1);
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"tz"),tzname[0],-1);

            /* create nice display time */
            t = time(NULL);
            str = ctime(&t);
            str[strlen(str) - 1] = '\0'; /* cut off newline */
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"display"),str,-1);
            deliver(dpacket_new(jp->x),NULL);
            return;
        }
	else if(NSCHECK(jp->iq,NS_VERSION))
        {
            log_debug(NAME, "[%s] Server packet - Version Request", FZONE);

            jutil_iqresult(jp->x);
            xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"),"xmlns",NS_VERSION);
            jpacket_reset(jp);

            xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"name"),NAME,-1);
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"version"),VERSION,-1);

            uname(&un);
            x = xmlnode_insert_tag(jp->iq,"os");
            xmlnode_insert_cdata(x,un.sysname,-1);
            xmlnode_insert_cdata(x," ",1);
            xmlnode_insert_cdata(x,un.release,-1);

            deliver(dpacket_new(jp->x),NULL);
            return;
        }
	else if(NSCHECK(jp->iq,NS_BROWSE))
        { 
            log_debug(NAME, "[%s] Server packet - Browse Request", FZONE);

            jutil_iqresult(jp->x);
            xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"conference"),"xmlns",NS_BROWSE);
            jpacket_reset(jp);

            xmlnode_put_attrib(jp->iq,"type","public");

            /* pull name from the server vCard */
            xmlnode_put_attrib(jp->iq,"name",xmlnode_get_tag_data(master->config,"vCard/FN")); 
	    xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "ns"), NS_MUC, -1);
	    xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "ns"), NS_REGISTER, -1);
	    xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "ns"), NS_VERSION, -1);
	    xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "ns"), NS_TIME, -1);
	    xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "ns"), NS_LAST, -1);

            xhash_walk(master->rooms, con_server_browsewalk, (void*)jp);

	    /*
	    x = xmlnode_insert_tag(jp->iq, "conference");
            xmlnode_put_attrib(x, "type", "public");
            xmlnode_put_attrib(x, "jid", "test@test");
	    */

            deliver(dpacket_new(jp->x),NULL);
            return;
        }
	else if(NSCHECK(jp->iq, NS_DISCO_INFO))
	{
            log_debug(NAME, "[%s] Server packet - Disco Request", FZONE);

            jutil_iqresult(jp->x);
            xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"), "xmlns", NS_DISCO_INFO);
            jpacket_reset(jp);

	    x = xmlnode_insert_tag(jp->iq,"identity");
	    xmlnode_put_attrib(x, "category", "conference");
	    xmlnode_put_attrib(x, "type", "text");
	    xmlnode_put_attrib(x, "name", xmlnode_get_tag_data(master->config, "vCard/FN"));

            xmlnode_put_attrib(xmlnode_insert_tag(jp->iq,"feature"), "type", NS_GROUPCHAT);
            xmlnode_put_attrib(xmlnode_insert_tag(jp->iq,"feature"), "type", NS_MUC);

            deliver(dpacket_new(jp->x),NULL);
            return;
	}
	else if(NSCHECK(jp->iq, NS_LAST))
        {
            log_debug(NAME, "[%s] Server packet - Last Request", FZONE);

            jutil_iqresult(jp->x);
            xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"),"xmlns",NS_LAST);
            jpacket_reset(jp);

            start = time(NULL) - master->start;
            sprintf(nstr,"%d",start);
            xmlnode_put_attrib(jp->iq,"seconds",nstr);

            deliver(dpacket_new(jp->x),NULL);
            return;
        }
	else if(NSCHECK(jp->iq,NS_VCARD))
        { 
            log_debug(NAME, "[%s] Server packet - VCard Request", FZONE);

            jutil_iqresult(jp->x);
            xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"vCard"),"xmlns",NS_VCARD);
            jpacket_reset(jp);

            xmlnode_insert_node(jp->iq,xmlnode_get_firstchild(xmlnode_get_tag(master->config,"vCard")));

            deliver(dpacket_new(jp->x),NULL);
            return;
        }
    }

    jutil_error(jp->x, TERROR_NOTIMPL);
    deliver(dpacket_new(jp->x),NULL);
}


void _con_packets(void *arg)
{
    jpacket jp = (jpacket)arg;
    cni master = (cni)jp->aux1;
    cnr room;
    cnu u, u2;
    char *s, *reason;
    xmlnode node;
    int priority = -1;
    int created = 0;

    /* first, handle all packets just to the server (browse, vcard, ping, etc) */
    if(jp->to->user == NULL)
    {
        con_server(master, jp);
        return;
    }

    log_debug(NAME, "[%s] processing packet %s", FZONE, xmlnode2str(jp->x));

    /* any other packets must have an associated room */
    for(s = jp->to->user; *s != '\0'; s++) 
	    *s = tolower(*s); /* lowercase the group name */

    if((room = xhash_get(master->rooms, jid_full(jid_user(jp->to)))) == NULL)
    {
	if(master->roomlock == 1)
	{
	    log_debug(NAME, "[%s] Room building request denied", FZONE);
	    jutil_error(jp->x, TERROR_MUC_ROOM);
	    deliver(dpacket_new(jp->x),NULL);
	    return;
	}
	else if(jp->type == JPACKET_IQ && jpacket_subtype(jp) == JPACKET__GET && NSCHECK(jp->iq, NS_MUC_OWNER))
        {
            room = con_room_new(master, jid_user(jp->to), jp->from, NULL, NULL, 1, 1, 0);
	    xmlnode_free(jp->x);
	    return;
	}
	else if(jp->to->resource == NULL)
	{
	    log_debug(NAME, "[%s] Room %s doesn't exist: Returning Bad Request", FZONE, jp->to->user);
	    jutil_error(jp->x, TERROR_BAD);
	    deliver(dpacket_new(jp->x),NULL);
	    return;
	}
	else if(jpacket_subtype(jp) == JPACKET__UNAVAILABLE)
	{
	    log_debug(NAME, "[%s] Room %s doesn't exist: dropping unavailable presence", FZONE, jp->to->user);
	    xmlnode_free(jp->x);
	    return;
	}
	else
	{
            room = con_room_new(master, jid_user(jp->to), jp->from, NULL, NULL, 1, 0, 0);
            /* fall through, so the presence goes to the room like normal */
	    created = 1;
	}
    }

    /* get the sending user entry, if any */
    u = xhash_get(room->remote, jid_full(jp->from));

    /* several things use this field below as a flag */
    if(jp->type == JPACKET_PRESENCE)
        priority = jutil_priority(jp->x);

    /* sending available presence will automatically get you a generic user, if you don't have one */
    if(u == NULL && priority >= 0)
        u = con_user_new(room, jp->from);

    /* update tracking stuff */
    room->last = time(NULL);
    room->packets++;

    if(u != NULL)
    {
        u->last = time(NULL);
        u->packets++;
    }

    /* handle the old-style conferencing join/rename */
    if(priority >= 0 && jp->to->resource != NULL)
    {
        u2 = con_room_usernick(room, jp->to->resource); /* existing user w/ this nick? */

        /* it's just us updating our presence, old fashioned way */
        if(u2 == u)
        {
            jp->to = jid_user(jp->to);
            xmlnode_put_attrib(jp->x,"to",jid_full(jp->to));
            con_room_process(room, u, jp);
            return;
        }

	/* User already exists, return conflict Error */
        if(u2 != NULL)
        {
            jutil_error(jp->x, TERROR_MUC_NICK);
            deliver(dpacket_new(jp->x),NULL);
            return;
        }

	/* Nick already registered, return conflict Error */
        if(is_registered(master, jid_full(jid_user(u->realid)), jp->to->resource) == -1)
        {
            jutil_error(jp->x, TERROR_MUC_NICKREG);
            deliver(dpacket_new(jp->x),NULL);
            return;
        }

        /* if from an existing user in the room, change the nick */
	if(is_outcast(room, u->realid) && !is_admin(room, u->realid))
	{
	    jutil_error(jp->x, TERROR_MUC_BANNED);
	    deliver(dpacket_new(jp->x),NULL);
	    return;
	}

	/* User is not invited, return invitation error */
	if(room->invitation == 1 && !is_member(room, u->realid) && !is_owner(room, u->realid))
	{
	    jutil_error(jp->x, TERROR_MUC_INVITED);
	    deliver(dpacket_new(jp->x),NULL);
	    return;
	}

	/* Room is full, return full room error */
	if(room->count >= room->maxusers && room->maxusers != 0 && !is_admin(room, u->realid))
	{
	    log_debug(NAME, "[%s] Room over quota - disallowing entry", FZONE);
	    jutil_error(jp->x, TERROR_MUC_FULL);
	    deliver(dpacket_new(jp->x),NULL);
	    return;
	}

	/* Room has been locked against entry */
	if(room->locked)
	{
	    log_debug(NAME, "[%s] Room has been locked", FZONE);
	    jutil_error(jp->x, TERROR_FORBIDDEN);
	    deliver(dpacket_new(jp->x),NULL);
	    return;
	}

	/* User already in room, simply a nick change */
        if(u->localid != NULL)
        {
            con_user_nick(u, jp->to->resource, NULL); /* broadcast nick rename */
	    xmlnode_free(jp->x);
	    return;

        }
	else if(room->secret == NULL) /* No password required, just go right in */
	{
	    if(NSCHECK(xmlnode_get_tag(jp->x,"x"), NS_MUC))
	    {
		/* Set legacy value to room value */
	        u->legacy = room->legacy;
		node = xmlnode_get_tag(jp->x,"x");
		xmlnode_hide(node);
	    }
	    else
	    {
		u->legacy = 1;
		created = 0; /* Override created flag for non-MUC compliant clients */
	    } 

	    u->presence = xmlnode_dup(jp->x);
	    jutil_delay(u->presence, NULL);

  	    log_debug(NAME, "[%s] About to enter room, legacy<%d>, presence [%s]", FZONE, u->legacy, xmlnode2str(u->presence));
            con_user_enter(u, jp->to->resource, created); /* join the room */

	    xmlnode_free(jp->x);
	    return;

        }
	else if(jp->type == JPACKET_PRESENCE) /* Hopefully you are including a password, this room is locked */
	{
	    if(NSCHECK(xmlnode_get_tag(jp->x,"x"), NS_MUC))
	    {
	        log_debug(NAME, "[%s] Password?", FZONE);
		if(j_strcmp(room->secret,xmlnode_get_tag_data(xmlnode_get_tag(jp->x,"x"),"password")) == 0)
		{
		    /* Set legacy value to room value */
		    u->legacy = room->legacy;
		    node = xmlnode_get_tag(jp->x,"x");
		    xmlnode_hide(node);

	            u->presence = xmlnode_dup(jp->x);
	            jutil_delay(u->presence, NULL);
		    con_user_enter(u, jp->to->resource, created); /* join the room */

		    xmlnode_free(jp->x);
		    return;
		}
	    }
	}

	/* No password found, room is password protected. Return password error */
        jutil_error(jp->x, TERROR_MUC_PASSWORD);
        deliver(dpacket_new(jp->x),NULL);
        return;
    }

    /* kill any user sending unavailable presence */
    if(jpacket_subtype(jp) == JPACKET__UNAVAILABLE)
    {
	reason = xmlnode_get_tag_data(jp->x, "status");

	node = xmlnode_new_tag("reason");
	if (reason)
	    xmlnode_insert_cdata(node, reason, -1);

        con_user_zap(u, node);
        xmlnode_free(jp->x);
        return;
    }

    /* handle errors */
    if(jpacket_subtype(jp) == JPACKET__ERROR)
    {
        /* only allow iq errors that are to a resource (direct-chat) */
        if(jp->to->resource == NULL || jp->type != JPACKET_IQ)
	{
	    node = xmlnode_new_tag("reason");
	    xmlnode_insert_cdata(node, "Lost connection", -1);

            con_user_zap(u, node);
	}

        xmlnode_free(jp->x);
        return;
    }

    /* not in the room yet? foo */
    if(u == NULL || u->localid == NULL)
    {
	if(u == NULL)
	{
		log_debug(NAME, "[%s] No cnu found for user", FZONE);
	}
	else
	{
		log_debug(NAME, "[%s] No lid found for %s", FZONE, jid_full(u->realid));
	}

        if(jp->to->resource != NULL)
        {
            jutil_error(jp->x, TERROR_NOTFOUND);
            deliver(dpacket_new(jp->x),NULL);
        }
	else
	{
            con_room_outsider(room, u, jp); /* non-participants get special treatment */
        }
	
        return;
    }

    /* packets to a specific resource?  one on one chats, browse lookups, etc */
    if(jp->to->resource != NULL)
    {
        if((u2 = xhash_get(room->local, jp->to->resource)) == NULL && (u2 = con_room_usernick(room, jp->to->resource)) == NULL) /* gotta have a recipient */
        {
            jutil_error(jp->x, TERROR_NOTFOUND);
            deliver(dpacket_new(jp->x),NULL);
        }
	else
	{
            con_user_process(u2, u, jp);
        }
        
	return;
    }

    /* finally, handle packets just to a room from a participant, msgs, pres, iq browse/conferencing, etc */
    con_room_process(room, u, jp);

}


/* phandler callback, send packets to another server */
result con_packets(instance i, dpacket dp, void *arg)
{
    cni master = (cni)arg;
    jpacket jp;

    /* routes are from dnsrv w/ the needed ip */
    if(dp->type == p_ROUTE)
        jp = jpacket_new(xmlnode_get_firstchild(dp->x));
    else
        jp = jpacket_new(dp->x);

    /* if the delivery failed */
    if(jp == NULL)
    {
        deliver_fail(dp,"Illegal Packet");
        return r_DONE;
    }

    /* bad packet??? ick */
    if(jp->type == JPACKET_UNKNOWN || jp->to == NULL)
    {
        jutil_error(jp->x, TERROR_BAD);
        deliver(dpacket_new(jp->x),NULL);
        return r_DONE;
    }

    /* we want things processed in order, and don't like re-entrancy! */
    jp->aux1 = (void*)master;
    mtq_send(master->q, jp->p, _con_packets, (void *)jp);
    return r_DONE;
}

/* callback for walking each user in a room */
void _con_beat_user(xht h, const char *key, void *data, void *arg)
{
    cnu user = (cnu)data;
    int now = (int)arg;
    xmlnode node;

    if(user->localid == NULL && (now - user->last) > 120)
    {
	node = xmlnode_new_tag("reason");
        xmlnode_insert_cdata(node, "Clearing zombie", -1);
        con_user_zap(user, node);
    }
}

/* callback for walking each room */
void _con_beat_idle(xht h, const char *key, void *data, void *arg)
{
    cnr room = (cnr)data;
    int now = (int)arg;

    xhash_walk(room->remote,_con_beat_user,arg); /* makes sure nothing stale is in the room */

    if(room->persistant == 0 && room->count == 0 && (now - room->last) > 240)
        con_room_zap(room);
}

/* heartbeat checker for timed out idle rooms */
void _con_beat_logrotate(xht h, const char *key, void *data, void *arg)
{
    cnr room = (cnr)data;

    if(room->logfile)
    {
	log_debug(NAME, "[%s] Rotating log for room %s", FZONE, jid_full(room->id));

	con_room_log_close(room);
	con_room_log_new(room);
    }
}

/* heartbeat checker for timed out idle rooms */
void _con_beat_logupdate(xht h, const char *key, void *data, void *arg)
{
    cnr room = (cnr)data;
    char *timestamp = (char*)arg;

    if(room->logformat == LOG_XHTML && room->logfile)
    {
	log_debug(NAME, "[%s] Adding anchor >%s< for room %s", FZONE, timestamp, jid_full(room->id));
	fprintf(room->logfile, "<a href=\"%s\">\n", timestamp);
	fflush(room->logfile);
    }
}

/* heartbeat checker for timed out idle rooms */
result con_beat_idle(void *arg)
{
    cni master = (cni)arg;

    xhash_walk(master->rooms,_con_beat_idle,(void*)time(NULL));
    return r_DONE;
}

/* heartbeat checker for log rotation */
result con_beat_logrotate(void *arg)
{
    cni master = (cni)arg;

    if(j_strcmp(master->day, dateget()) == 0)
        return r_DONE;

    master->day = dateget();

    xhash_walk(master->rooms,_con_beat_logrotate, NULL);
    return r_DONE;
}

/* heartbeat checker for timed out idle rooms */
result con_beat_logupdate(void *arg)
{
    cni master = (cni)arg;
    int mins = minuteget();

    if(mins % 5 == 0)
    {
        xhash_walk(master->rooms,_con_beat_logupdate,(void*)timeget());
    }

    return r_DONE;
}

/* heartbeat checker for miscellaneous tasks */
result con_beat_housekeep(void *arg)
{
    cni master = (cni)arg;

    /* Check if rooms have been loaded */
    if(master->loader == 0)
    {
	master->loader = 1;
	xdb_rooms_get(master);
    }

    return r_DONE;
}

/*** everything starts here ***/
void conference(instance i, xmlnode x)
{
    cni master;
    xmlnode cfg;
    jid sadmin;

    log_debug(NAME, "[%s] mu-conference loading", FZONE);

    /* Allocate space for cni struct and link to instance */
    master = pmalloco(i->p,sizeof(_cni));
    master->i = i;

    /* Set up xdb interface */
    master->xdbc = xdb_cache(i);

    /* get the config */
    cfg = xdb_get(master->xdbc, jid_new(xmlnode_pool(x), "config@-internal"), "jabber:config:conference");

    /* Parse config and initialise variables */
    master->q = mtq_new(i->p);
    master->maxhash = j_atoi(xmlnode_get_tag_data(cfg,"maxhash"),401);
    master->loader = 0;
    master->rooms = xhash_new(j_atoi(xmlnode_get_tag_data(cfg,"maxrooms"),master->maxhash));
    master->history = j_atoi(xmlnode_get_tag_data(cfg,"history"),10);
    master->config = cfg;					/* Store a copy of the config for later usage */
    master->day = dateget();					/* Used to determine when to rotate logs */
    master->logdir = xmlnode_get_tag_data(cfg, "logdir");	/* Directory where to store logs */

    /* If requested, set default room state to 'public', otherwise will default to 'private */
    if(xmlnode_get_tag(cfg,"public"))
        master->public = 1;

    /* If requested, stop any new rooms being created */
    if(xmlnode_get_tag(cfg,"roomlock"))
        master->roomlock = 1;

    /* If requested, stop any new rooms being created */
    if(xmlnode_get_tag(cfg,"dynamic"))
        master->dynamic = 1;

    /* sadmin code */
    if(xmlnode_get_tag(cfg, "sadmin"))
    {
	sadmin = jid_new(xmlnode_pool(master->config), xmlnode_get_tag_data(cfg, "sadmin"));
	master->sadmin = jid_user(sadmin);
    }

    register_phandler(i,o_DELIVER,con_packets,(void*)master);
    register_beat(120, con_beat_idle, (void *)master);
    register_beat(60, con_beat_logrotate, (void *)master);
    register_beat(60, con_beat_logupdate, (void *)master);
    register_beat(5, con_beat_housekeep, (void *)master);

}
